import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.ImageObserver;

/**
 * Utility class for manipulating BufferedImages.
 */
public class BufferedImages {
	private BufferedImages() {}
	
	/**
	 * Returns an array of integer pixels in the default RGB color model
	 * (BufferedImage.TYPE_INT_ARGB) and default sRGB color space that
	 * comprise the entire image.
	 * 
	 * This method is equivalent to calling
	 * <blockquote>
	 *     <tt>BufferedImage.getRGB(0, 0, img.getWidth(), img.getHeight(), argbArray, 0, img.getWidth())</tt>
	 * </blockquote>
	 * except that it is faster when <tt>img</tt> is already in
	 * the native ARGB color model.
	 * 
	 * The <tt>offset</tt> of the output array is assumed to be <tt>0</tt>, and
	 * <tt>scansize</tt> is assumed to be <tt>img.getWidth()</tt>.
	 * 
	 * With a specific coordinate (x, y) in the image, the ARGB pixel can
	 * be accessed in this way:
	 * <blockquote>
	 *     <tt>pixel = argbArray[y*img.getWidth() + x]</tt>
	 * </blockquote>
	 * 
	 * @param img		the image whose data is to be extracted
	 * @param argbArray	if not <tt>null</tt> and is of the correct length,
	 * 					the pixels are written here
	 * @return			array of ARGB pixels
	 * @see {@link java.awt.image.BufferedImage#getRGB(int,int,int,int,int[],int,int)}
	 */
	public static int[] getAllRGB(BufferedImage img, int[] argbArray) {
		/* Try to use optimized ARGB access strategy if the image is in the correct format */
		if (img.getColorModel().equals(ColorModel.getRGBdefault())) {
			DataBuffer dataBuf = img.getRaster().getDataBuffer();
			int numPixels = dataBuf.getSize();
			
			// Ensure output pixel array is allocated and is of the correct length
			if ((argbArray == null) || (argbArray.length != numPixels)) {
				argbArray = new int[numPixels];
			}
			
			// Copy all packed ARGB pixels (in the sRGB color model) from
			// data buffer to the output pixel array
			for (int i=0; i<numPixels; i++) {
				argbArray[i] = dataBuf.getElem(i);
			}
			
			return argbArray;
		}
		
		/* Fallback to using general ARGB access strategy */
		return img.getRGB(0, 0, img.getWidth(), img.getHeight(), argbArray, 0, img.getWidth());
	}
	
	/**
	 * Sets an array of integer pixels in the default RGB color model
	 * (BufferedImage.TYPE_INT_ARGB) and default sRGB color space to
	 * the set of pixels that comprise the entire image.
	 * 
	 * This method is equivalent to calling
	 * <blockquote>
	 *     <tt>BufferedImage.setRGB(0, 0, img.getWidth(), img.getHeight(), argbArray, 0, img.getWidth())</tt>
	 * </blockquote>
	 * except that it is faster when <tt>img</tt> is already in
	 * the native ARGB color model.
	 * 
	 * The <tt>offset</tt> of the input array is assumed to be <tt>0</tt>, and
	 * <tt>scansize</tt> is assumed to be <tt>img.getWidth()</tt>.
	 * 
	 * With a specific coordinate (x, y) in the image, the ARGB pixel can
	 * be accessed in this way:
	 * <blockquote>
	 *     <tt>pixel = argbArray[y*img.getWidth() + x]</tt>
	 * </blockquote>
	 * 
	 * @param img		the image whose data is to be changed
	 * @param argbArray	the ARGB pixels
	 * @see {@link java.awt.image.BufferedImage#setRGB(int,int,int,int,int[],int,int)}
	 */
	public static void setAllRGB(BufferedImage img, int[] argbArray) {
		/* Try to use optimized ARGB access strategy if the image is in the correct format */
		if (img.getColorModel().equals(ColorModel.getRGBdefault())) {
			DataBuffer dataBuf = img.getRaster().getDataBuffer();
			int numPixels = dataBuf.getSize();
			
			// Ensure input pixel array is of the expected size
			if (argbArray.length != numPixels)
				throw new IllegalArgumentException(
					"expected argbArray to have "+numPixels+" element(s)");
			
			// Copy all packed ARGB pixels (in the sRGB color model) from
			// the pixel array to the data buffer
			for (int i=0; i<numPixels; i++) {
				dataBuf.setElem(i, argbArray[i]);
			}
			
			return;
		}
		
		/* Fallback to using general ARGB access strategy */
		img.setRGB(0, 0, img.getWidth(), img.getHeight(), argbArray, 0, img.getWidth());
	}
	
	/*
	public static void setAllGrays(BufferedImage img, int[] grayArray) {
		// Try to use optimized gray access strategy if the image is in the correct format
		if (img.getType() == BufferedImage.TYPE_BYTE_GRAY) {
			DataBuffer dataBuf = img.getRaster().getDataBuffer();
			int numPixels = dataBuf.getSize();
			
			// Ensure input pixel array is of the expected size
			if (grayArray.length != numPixels)
				throw new IllegalArgumentException(
					"expected grayArray to have "+numPixels+" element(s)");
			
			// Copy all gray pixels from the pixel array to the data buffer
			for (int i=0; i<numPixels; i++) {
				dataBuf.setElem(i, grayArray[i]);
			}
			
			return;
		}
		
		// XXX: implement
		throw new UnsupportedOperationException(
			"not implemented: grayscale-style access to non-grayscale images");
	}
	*/
	
	/**
	 * Like {@link #createSizeCompatibleBufferedImage(Image,BufferedImage,int)},
	 * but always returns a BufferedImage that uses the default RGB color model
	 * (BufferedImage.TYPE_INT_ARGB).
	 */
	public static BufferedImage createSizeCompatibleARGBBufferedImage(
		Image inImg,
		BufferedImage outBufImg)
	{
		return createSizeCompatibleBufferedImage(inImg, outBufImg, BufferedImage.TYPE_INT_ARGB);
	}
	
	/**
	 * @param inImg		the image that the returned BufferedImage must be size-compatible with
	 * @param outBufImg	if not <tt>null</tt> and is of size matching <tt>inImg</tt> and type
	 * 					matching <tt>outBufImgType</tt>, will be returned as the output image
	 * @return			a BufferedImage that is size-compatible with <tt>inImg</tt>
	 */
	public static BufferedImage createSizeCompatibleBufferedImage(
		Image inImg,
		BufferedImage outBufImg,
		int outBufImgType)
	{
		int width = inImg.getWidth(null);
		int height = inImg.getHeight(null);
		
		if ((outBufImg == null) ||
			(outBufImg.getWidth() != width) ||
			(outBufImg.getHeight() != height) ||
			(outBufImg.getType() != outBufImgType))
		{
			outBufImg = new BufferedImage(
				width,
				height,
				outBufImgType);
		}
		
		return outBufImg;
	}
	
	public static BufferedImage copyImageToARGBBufferedImage(
		Image srcImg, BufferedImage dstImg)
	{
		return copyImageToBufferedImage(srcImg, dstImg, BufferedImage.TYPE_INT_ARGB);
	}
	
	/**
	 * Copies the pixels from <tt>srcImg</tt> into <tt>dstImg</tt> (creating a
	 * destination image if <tt>dstImg</tt> is <tt>null</tt>) and returns
	 * the destination image.
	 * 
	 * @param srcImg	the image whose pixels are to be copied
	 * @param dstImg	if not <tt>null</tt>, the pixels from <tt>srcImg</tt> are copied to here
	 * @return			the destination image (into which the <tt>srcImg</tt>'s pixels were copied)
	 */
	public static BufferedImage copyImageToBufferedImage(
		Image srcImg, BufferedImage dstImg, int dstImgType)
	{
		// Ensure dstImg is non-null and size-compatible
		dstImg = BufferedImages.createSizeCompatibleBufferedImage(srcImg, dstImg, dstImgType);
		
		// Copy the data from the original image into the BufferedImage
		dstImg.getGraphics().drawImage(srcImg, 0, 0, (ImageObserver) null);
		
		return dstImg;
	}
	
	/**
	 * @param inImg				the image that the returned pixel buffer must be size-compatible with
	 * @param outPixelBuffer	if not <tt>null</tt> and is of the correct size, will be returned
	 * @return					a pixel buffer that is size-compatible with <tt>inImg</tt>;
	 * 							allocated assuming that each pixel of an image occupies exactly
	 * 							one element of the pixel buffer
	 */
	public static int[] createSizeCompatiblePixelBuffer(Image inImg, int[] outPixelBuffer) {
		int width = inImg.getWidth(null);
		int height = inImg.getHeight(null);
		int numPixels = width*height;
		
		// Ensure the size of the pixel buffer matches the number of pixels in the image
		if ((outPixelBuffer == null) || (outPixelBuffer.length != numPixels)) {
			outPixelBuffer = new int[numPixels];
		}
		
		return outPixelBuffer;
	}
	
	/**
	 * Copies the pixels from <tt>srcImg</tt> into <tt>dstPixelBuffer</tt> (creating a
	 * destination pixel buffer if <tt>dstPixelBuffer</tt> is <tt>null</tt>) and returns
	 * the destination pixel buffer.
	 * 
	 * @param srcImg			the image whose pixels are to be copied
	 * @param dstPixelBuffer	if not <tt>null</tt>, the pixels from <tt>srcImg</tt> are copied to here
	 * @return					the destination pixel buffer (into which the <tt>srcImg</tt>'s pixels were copied)
	 */
	public static int[] copyBufferedImageToARGBPixelBuffer(BufferedImage srcImg, int[] dstPixelBuffer) {
		// Ensure the size of the output pixel buffer matches the number of pixels in the input image
		dstPixelBuffer = BufferedImages.createSizeCompatiblePixelBuffer(srcImg, dstPixelBuffer);
		
		// Copy the image's pixels into the pixel buffer
		BufferedImages.getAllRGB(srcImg, dstPixelBuffer);
		
		return dstPixelBuffer;
	}
}