import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import javax.swing.*;

public class LinearHoughTransform2 implements ImageFilter {
	// ### FILTER ###
	
	/* First-time initialized */
	private static double[] sinTable = null;	// indexed by degrees; of length 360
	private static double[] cosTable = null;	// indexed by degrees; of length 360
	
	/* Reinitialized upon change of image size */
	private int lastWidth = -1;
	private int lastHeight = -1;
	private int originX = -1;
	private int originY = -1;
	private int maxLineLen = -1;
	private int[] sumForLine = null;			// 2D matrix [deg,r] with the sums of the grays along the corresponding line
	private BufferedImage outImage = null;
	
	/* Every-time initialized */
	private BufferedImage inImgBuffer = null;
	private int[] inPixelBuffer = null;
	private int[] inGrayPixels;
	
	public Image createFilteredImage(Image img) {
		int width = img.getWidth(null);
		int height = img.getHeight(null);
		
		/*
		long DEBUG_startTime = System.currentTimeMillis();			// DEBUG: remove
		*/
		
		/* First-time initialization */
		if (this.sinTable == null) {
			/* Compute sinTable and cosTable */
			this.sinTable = new double[360];
			this.cosTable = new double[360];
			for (int deg=0; deg<360; deg++) {
				double rad = deg*StrictMath.PI/180.0;
				this.sinTable[deg] = StrictMath.sin(rad);
				this.cosTable[deg] = StrictMath.cos(rad);
			}
		}
		
		/* Initialization upon change of image size */
		if ((width != lastWidth) || (height != lastHeight)) {
			this.originX = width/2;
			this.originY = height/2;
			
			this.maxLineLen = (((int) StrictMath.sqrt(width*width + height*height)) + 1)/2 + 1;		// +1's: compensate for roundoff error
			
			int sumForLine_numRs = 2*maxLineLen + 1;
			int sumForLine_numThetas = 180;
			this.sumForLine = new int[sumForLine_numRs*sumForLine_numThetas];
			
			// TODO: rebuild output image buffer
			this.outImage = new BufferedImage(
				sumForLine_numRs, sumForLine_numThetas,
				BufferedImage.TYPE_BYTE_GRAY);
			
			this.lastWidth = width;
			this.lastHeight = height;
		}
		
		/* Every-time initialization */
		this.inImgBuffer = BufferedImages.copyImageToARGBBufferedImage(img, this.inImgBuffer);
		this.inPixelBuffer = BufferedImages.copyBufferedImageToARGBPixelBuffer(inImgBuffer, this.inPixelBuffer);
		this.inGrayPixels = Pixels.ARGBPixelsToGrayPixels(inPixelBuffer, this.inGrayPixels);
		
		/* Iterate over lines in image, summing up the gray levels along each one */
		for (int thetaDeg=0, phiDeg=90, sumForLine_off=0;
		     thetaDeg<180;
		     thetaDeg++, phiDeg++)
		{
			for (int r = -maxLineLen; r<=maxLineLen; r++, sumForLine_off++) {
				double rCosThetaDeg = r*cosTable[thetaDeg];	// r*cos(thetaDeg)
				double rSinThetaDeg = r*sinTable[thetaDeg];	// r*sin(thetaDeg)
				
				/* Iterate over points in line, summing up the gray values */
				int sumOfGraysInLine = 0;
				for (int pos = -maxLineLen; pos<=maxLineLen; pos++) {
					double posCosPhiDeg = pos*cosTable[phiDeg];	// pos*cos(phiDeg)
					double posSinPhiDeg = pos*sinTable[phiDeg];	// pos*sin(phiDeg)
					
					int x = originX + (int) (rCosThetaDeg + posCosPhiDeg);	// originX + r*cos(thetaDeg) + pos*cos(phiDeg)
					int y = originY + (int) (rSinThetaDeg + posSinPhiDeg);	// originY + r*sin(thetaDeg) + pos*sin(phiDeg)
					
					if ((x >= 0) && (x < width) && (y >= 0) && (y < height)) {	// bound check
						sumOfGraysInLine += inGrayPixels[y*width + x];
					}
				}
				
				// sumForLine[thetaDeg,r] = sumOfGraysInLine
				sumForLine[sumForLine_off] = sumOfGraysInLine;
			}
		}
		
		/* Normalize the sums to fall in the range [0,255] */
		{
			int maxSum = 0;
			for (int i=0, n=sumForLine.length; i<n; i++) {
				int sumForCurLine = sumForLine[i];
				if (sumForCurLine > maxSum) {
					maxSum = sumForCurLine;
				}
			}
			
			if (maxSum > 0) {
				for (int i=0, n=sumForLine.length; i<n; i++) {
					sumForLine[i] = sumForLine[i]*255/maxSum;
				}
			}
		}
		
		/* Copy sumForLine data into the output image */
		DataBuffer outImageBuffer = outImage.getRaster().getDataBuffer();
		for (int i=0, n=sumForLine.length; i<n; i++) {
			outImageBuffer.setElem(i, sumForLine[i]);
		}
		
		/*
		long DEBUG_endTime = System.currentTimeMillis();			// DEBUG: remove
		long DEBUG_deltaTime = (DEBUG_endTime-DEBUG_startTime);
		System.out.println("DEBUG: "+DEBUG_deltaTime);
		*/
		
		return outImage;
	}
	
	public String toString() {
		return "Linear Hough Transform #2";
	}
}